@@ -6,14 +6,20 @@
 import copy 
 import json
 import sqlite3
+import datetime
 import multiprocessing
 from typing import Any, List
-from . import lib, cursor, query
+from urllib.parse import urlparse
+from . import adapters, lib, cursor, jsonquery
+
+def _json_default(o):
+    if type(o) is datetime.date or type(o) is datetime.datetime:
+        return o.isoformat()
 
 
-def _to_json(data:dict) -> str:
+def json_dumps(data:dict) -> str:
     """
-    Convert data to json string
+    Convert data:dict to json string
 
     Args:
         data: dict
@@ -22,12 +28,12 @@ def _to_json(data:dict) -> str:
         str
 
     """
-    return json.dumps(data)
+    return json.dumps(data, default=_json_default)
 
 
-def _from_json(data:str) -> dict:
+def json_loads(data:str) -> dict:
     """
-    Convert data to json string
+    Convert data:str to json string
 
     Args:
         data: str
@@ -39,23 +45,26 @@ def _from_json(data:str) -> dict:
     return json.loads(data)
 
 
-def _row_to_dict(row:sqlite3.Row) -> dict:
+def parse_row_to_dict(row:dict) -> dict:
     """
-    Properly parse the SQLite#Row to document dict 
-
+    Convert a result row to dict, by merging _json with the rest of the columns
+    
     Args:
-        row: sqlite.Row
+        row: dict
 
     Returns
         dict
     """
-    _json = row["_json"]
-    doc = {**_from_json(_json)}
-    doc.update({k: row[k] for k in row.keys() if k not in ["_json"]})
-    return doc
+    row = row.copy()
+    _json = json_loads(row.pop("_json")) if "_json" in row else {}
+    return {
+        **_json,
+        **row # columns takes precedence. So whatever is there to overwrite _json
+    }
 
 
-def _parse_smart_filtering(filters: dict, indexed_columns:list=[]) -> dict:
+
+def parse_smart_filtering(filters: dict, indexed_columns:list=[]) -> dict:
     """
     Smart Filter
     Breaks down the filters based on type. 
@@ -80,14 +89,14 @@ def _parse_smart_filtering(filters: dict, indexed_columns:list=[]) -> dict:
     sql_filters = []
     json_filters = {}
     for k, v in filters.items():
-        if query.FILTER_OPERATOR in k:
-            f, op = k.split(query.FILTER_OPERATOR)
-            if f in indexed_columns and op in query.SQL_OPERATORS:
-                sql_filters.append((f, query.SQL_OPERATORS[op], v))
+        if jsonquery.FILTER_OPERATOR in k:
+            f, op = k.split(jsonquery.FILTER_OPERATOR)
+            if f in indexed_columns and op in jsonquery.SQL_OPERATORS:
+                sql_filters.append((f, jsonquery.SQL_OPERATORS[op], v))
                 continue
         else:
             if k in indexed_columns:
-                sql_filters.append((k, query.SQL_OPERATORS["eq"], v))
+                sql_filters.append((k, jsonquery.SQL_OPERATORS["eq"], v))
                 continue
         json_filters[k] = v
               
@@ -99,19 +108,32 @@ def _parse_smart_filtering(filters: dict, indexed_columns:list=[]) -> dict:
 
 class Database(object):
     """
-    ::Database
+    :: Database
     """
 
-    def __init__(self, filename=":memory:", search_proc=4):
+    def __init__(self, dsn:str=None, connector:adapters.Adapter = None):
         """
         Connect database 
 
         Return:
+            dsn: str
+                *`scheme://hostname` should be there at minimum
+                - sqlite://
+                - sqlite://./file.db
+                - mysql://username:password@hostname:port/dbname
+                - mariadb://username:password@hostname:port/dbname
         """
-        self.conn = sqlite3.connect(filename, isolation_level=None)
-        self.conn.execute('pragma journal_mode=wal')
-        self.conn.row_factory = sqlite3.Row
-        self._search_proc = search_proc
+        if not dsn and not connector:
+            connector = adapters.SQLiteAdapter(file=":memory:")
+        elif dsn:
+            _ = urlparse(dsn)
+            # scheme, username, password, hostname, port, path.strip("/")
+            if _.scheme == "sqlite":
+                connector = adapters.SQLiteAdapter(dsn)
+            elif _.scheme in ["mysql", "mariadb"]:
+                connector = adapters.MySQLAdapter(dsn)
+
+        self.conn = connector
 
     def select(self, name):
         """
@@ -120,7 +142,7 @@ class Database(object):
         Returns:
             Collection
         """
-        return Collection(self, name)
+        return Collection(self.conn, name)
 
     def __getattr__(self, __name: str):
         """
@@ -143,32 +165,23 @@ class Database(object):
         Returns: 
           List
         """
-        with self.conn:
-            q = "SELECT name FROM sqlite_master WHERE type=?"
-            _exec = self.conn.execute(q, ('table', ))
-            return [item["name"] for item in _exec.fetchall()]
-
-    def sync(self):
-        pass
+        return self.conn.get_collections()
 
 class Document(dict):
     """
     :: Document
     """
 
-    # Private columns
-    PRIVATE_COLUMNS = ["_id", "_json", "_created_at", "_modified_at"]
-
-    def __init__(self, collection, row: sqlite3.Row):
+    def __init__(self, collection, row:dict):
         self.collection = collection 
         self._load(row)
 
-    def dget(self, path:str, default=None)->Any:
+    def get(self, path:str, default=None)->Any:
         """
-        Return a property by DotNotation
+        Return a property by key/DotNotation
 
         ie: 
-            #dget("key.deep1.deep2.deep3")
+            #get("key.deep1.deep2.deep3")
 
         Args:
             path:str - the dotnotation path
@@ -179,9 +192,9 @@ class Document(dict):
         """
         return lib.dict_get(obj=dict(self), path=path, default=default)
 
-    def dset(self, path:str, value:Any):
+    def set(self, path:str, value:Any):
         """
-        Set a property by DotNotation
+        Set a property by key/DotNotation
 
         Args:
             path:str - the dotnotation path
@@ -194,9 +207,9 @@ class Document(dict):
         lib.dict_set(obj=data, path=path, value=value)
         self.update(data)
 
-    def dpop(self, path:str):
+    def pop(self, path:str):
         """ 
-        Remove a property by DotNotation and return the value
+        Remove a property by key/DotNotation and return the value
 
         Args:
             path:str
@@ -205,7 +218,7 @@ class Document(dict):
             Any: the value that was removed
         """
         data = copy.deepcopy(dict(self))
-        v = lib.dict_pop(obj=data, path=path)
+        v = lib.dict_pop(obj=data, path=path) if "." in path else data.pop(path)
         self.update(data)
         return v
 
@@ -252,23 +265,17 @@ class Document(dict):
         data = dict(self)
         self.update(data)
 
-    def _load(self, row:sqlite3.Row):
+    def _load(self, row:dict):
         """
         load the content into the document
         
         Args:
-            row: sqlite3.Row
+            row: dict
         """
-
         self._empty_self()
-        row = {k: row[k] for k in row.keys()}
-        _json = _from_json(row.pop("_json"))
+        row = parse_row_to_dict(row)
         self._id = row.get("_id")
-        doc = {
-            **_json,
-            **row # row columns takes precendece
-        }
-        super().__init__(doc)
+        super().__init__(row)
 
     def _empty_self(self):
         """ clearout all properties """
@@ -285,22 +292,10 @@ class Collection(object):
     _columns = []
     _indexes = []
 
-    def __init__(self, db, name):
+    def __init__(self, conn:adapters.Adapter, name):
         self.name = name
-        self.db = db.conn
-        self.conn = db.conn
-        self._create()
-
-    def _create(self):
-        """
-        Create a table/collection
-
-        Returns:
-          None
-        """
-        with self.db:
-            q = "CREATE TABLE IF NOT EXISTS %s (_id VARCHAR(32) PRIMARY KEY, _json TEXT, _created_at TIMESTAMP, _modified_at TIMESTAMP)" % self.name
-            self.db.execute(q)
+        self.conn = conn
+        self.conn.create_collection(self.name)
 
     # ---- properties ----
 
@@ -312,13 +307,7 @@ class Collection(object):
         Returns:
             list
         """
-        if not self._columns:
-            with self.db:
-                try:
-                    _exec = self.db.execute("PRAGMA table_info(%s)" % self.name)
-                    self._columns = [r["name"] for r in _exec.fetchall()]
-                except: pass
-        return self._columns
+        return self.conn.get_columns(self.name)
 
     @property
     def indexes(self) -> list:
@@ -328,19 +317,7 @@ class Collection(object):
         Returns:
             list
         """
-        if not self._indexes:
-            with self.db:
-                try:
-                    indexes = []
-                    _exec = self.db.execute("PRAGMA index_list(%s)" % self.name)
-                    for item in _exec.fetchall():
-                        name = item["name"]
-                        _i2 = self.db.execute("PRAGMA index_info(%s)" % name)
-                        f = _i2.fetchone()
-                        indexes.append(f["name"])
-                    self._indexes = indexes
-                except: pass
-        return self._indexes
+        return self.conn.get_indexes(self.name)
 
     @property
     def size(self) -> int:
@@ -350,9 +327,7 @@ class Collection(object):
         Returns:
             int
         """
-        with self.db:
-            _exec = self.db.execute("SELECT count(*) FROM %s;" % self.name)
-            return _exec.fetchone()[0]
+        return self.conn.get_size(self.name)
 
     # ---- methods ----
 
@@ -387,8 +362,8 @@ class Collection(object):
         # If there is more search properties, take it to the json
         xparams = []
         xquery = []
-        smart_filters = _parse_smart_filtering(filters, indexed_columns=self.columns)
-        
+        smart_filters = parse_smart_filtering(filters, indexed_columns=self.columns)
+
         # Build the SQL query
         query = "SELECT * FROM %s " % self.name
 
@@ -404,14 +379,11 @@ class Collection(object):
         if xquery and xparams:
             query += " WHERE %s " % " AND ".join(xquery)
 
-        query += " LIMIT 1"
-
+        query += " LIMIT 1"       
         xparams = list(filters.values())
-        with self.db:
-            _exec = self.db.execute(query, xparams)
-            row = _exec.fetchone()
-            if row:
-                return Document(self, row) if _as_document else row
+        row = self.conn.fetchone(query, xparams)
+        if row:
+            return Document(self, row) if _as_document else row
         return None
 
     def insert(self, doc: dict, _as_document:bool=True) -> Document:
@@ -429,26 +401,25 @@ class Collection(object):
         if not isinstance(doc, dict):
             raise TypeError('Invalid data type. Must be a dict')
 
-        with self.db:
-            _id = lib.gen_id()
-            ts = lib.get_timestamp()
-            xcolumns = self.DEFAULT_COLUMNS[:]
-            xparams = [_id, _to_json(doc), ts, ts]
-            q = "INSERT INTO %s " % self.name
-            
-            # indexed data
-            # some data can't be overriden 
-            for col in self.columns:
-                if col in doc and col not in xcolumns:
-                    _data = doc[col]
-                    if _data:
-                        xcolumns.append(col)
-                        xparams.append(_data)
+        _id = lib.gen_id()
+        ts = lib.get_timestamp()
+        xcolumns = self.DEFAULT_COLUMNS[:]
+        xparams = [_id, json_dumps(doc), ts, ts]
+        q = "INSERT INTO %s " % self.name
+        
+        # indexed data
+        # some data can't be overriden 
+        for col in self.columns:
+            if col in doc and col not in xcolumns:
+                _data = doc[col]
+                if _data:
+                    xcolumns.append(col)
+                    xparams.append(_data)
 
-            q += " ( %s ) VALUES ( %s ) " % (",".join(xcolumns), ",".join(["?" for _ in xparams]))
-            
-            self.db.execute(q, xparams)
-            return self.get(_id=_id, _as_document=_as_document)
+        q += " ( %s ) VALUES ( %s ) " % (",".join(xcolumns), ",".join(["?" for _ in xparams]))
+        
+        self.conn.execute(q, xparams)
+        return self.get(_id=_id, _as_document=_as_document)
 
     def update(self, _id: str, doc: dict = {}, replace: bool = False, _as_document=True) -> Document:
         """
@@ -463,34 +434,30 @@ class Collection(object):
         Returns:
             Document
         """
-        with self.db:
-            rdoc = self.get(_id=_id, _as_document=False)
-            if rdoc:
-                _doc = doc if replace else lib.dict_merge(_from_json(rdoc["_json"]), doc)
-                ts = lib.get_timestamp()
-                
-                if "_id" in _doc:
-                    del doc["_id"]
+        rdoc = self.get(_id=_id, _as_document=False)
+        if rdoc:
+            _doc = doc if replace else lib.dict_merge(json_loads(rdoc["_json"]), doc)
+            ts = lib.get_timestamp()
+            
+            xcolumns = ["_json", "_modified_at"]
+            xparams = [json_dumps(_doc), ts]
 
-                xcolumns = ["_json", "_modified_at"]
-                xparams = [_to_json(_doc), ts]
-
-                q = "UPDATE %s SET " % self.name
-                
-                # indexed data
-                # some data can't be overriden 
-                for col in self.columns:
-                    if col in _doc and col not in xcolumns:
-                        _data = _doc[col]
-                        if _data:
-                            xcolumns.append(col)
-                            xparams.append(_data)
-                q += ",".join(["%s = ?" % _ for _ in xcolumns])
-                q += " WHERE _id=?"
-                xparams.append(_id)
-                _exec = self.db.execute(q, xparams)
-                return self.get(_id=_id, _as_document=_as_document)
-            return None
+            q = "UPDATE %s SET " % self.name
+            
+            # indexed data
+            # some data can't be overriden 
+            for col in self.columns:
+                if col in _doc and col not in xcolumns:
+                    _data = _doc[col]
+                    if _data:
+                        xcolumns.append(col)
+                        xparams.append(_data)
+            q += ",".join(["%s = ?" % _ for _ in xcolumns])
+            q += " WHERE _id=?"
+            xparams.append(_id)
+            self.conn.execute(q, xparams)
+            return self.get(_id=_id, _as_document=_as_document)
+        return None
 
     def delete(self, _id: str) -> bool:
         """
@@ -502,8 +469,7 @@ class Collection(object):
         Returns:
             Bool
         """
-        with self.db:
-            self.db.execute("DELETE FROM %s WHERE _id=?" % (self.name), (_id, ))
+        self.conn.execute("DELETE FROM %s WHERE _id=?" % (self.name), (_id, ))
         return True
 
     def find(self, filters: dict = {}, sort: list = [], limit: int = 10, skip: int = 0) -> cursor.Cursor:
@@ -528,7 +494,8 @@ class Collection(object):
         # If there is more search properties, take it to the json
         xparams = []
         xquery = []
-        smart_filters = _parse_smart_filtering(filters, indexed_columns=self.columns)
+
+        smart_filters = parse_smart_filtering(filters, indexed_columns=self.columns)
         
         # Build the SQL query
         query = "SELECT * FROM %s " % self.name
@@ -548,19 +515,17 @@ class Collection(object):
         # Perform JSON search, as we have JSON_FILTERS
         # Full table scan, relative to WHERE clause
         chunk = 100
+        data = []
         if smart_filters["JSON_FILTERS"]:
-            _exec = self.db.execute(query, xparams)
-            data = []
-            while True:
-                chunked = _exec.fetchmany(chunk)
+            for chunked in self.conn.fetchmany(query, xparams, chunk):
                 if chunked:
-                    rows = [_row_to_dict(row) for row in chunked]
-                    for r in query.execute(rows, smart_filters["JSON_FILTERS"]):
+                    rows = [parse_row_to_dict(row) for row in chunked]
+                    for r in jsonquery.execute(rows, smart_filters["JSON_FILTERS"]):
                         data.append(r)
                 else:
                     break
             if data:
-                data = [Document(self, l1) for l1 in data]
+                data = [Document(self, d) for d in data]
             return cursor.Cursor(data, sort=sort, limit=limit, skip=skip)
 
         # Skip JSON SEARCH, use only SQL.
@@ -578,8 +543,8 @@ class Collection(object):
                 xparams.append(skip or 0)
                 xparams.append(limit or 10)
 
-            _exec = self.db.execute(query, xparams)            
-            data = [Document(self, _row_to_dict(row)) for row in _exec.fetchall()]
+            res = self.conn.fetchall(query, xparams)            
+            data = [Document(self, row) for row in res]
             return cursor.Cursor(data)
 
     def drop(self):
@@ -588,10 +553,8 @@ class Collection(object):
 
         Returns:
             None
-
         """
-        with self.db:
-            self.db.execute("DROP TABLE %s " % self.name)
+        self.conn.execute("DROP TABLE %s " % self.name)
 
     def add_columns(self, columns:List[str], enforce_index=False):
         """
@@ -627,34 +590,8 @@ class Collection(object):
                 if enforce_index and indx != "UNIQUE":
                     indx = True
                 cols_stmt.append((col, _type or "TEXT", indx))
-
-
-        with self.db:
-            for _ in cols_stmt:
-                try:
- 
-                    col, coltype, colindex = _
-
-                    # Add column if not exists
-                    if col not in self.columns: 
-                        self.db.execute("ALTER TABLE %s ADD COLUMN %s %s DEFAULT NULL" % (self.name, col, coltype))
-
-                    # Add index if not exists
-                    if colindex and col not in self.indexes:
-                        index_stmt = None
-                        # unique
-                        if isinstance(colindex, str) and colindex.upper() == "UNIQUE":
-                            index_stmt = "UNIQUE"
-                        # regular
-                        elif colindex is True:
-                            index_stmt = ""
-                        if index_stmt is not None:
-                            self.db.execute("CREATE %s INDEX %s ON %s (%s)" % (index_stmt, col+"__idx", self.name, col))
-                except Exception as e: pass
-        # reset columns
-        self._columns = []
-        self._indexes = []
-
+        self.conn.add_columns(table=self.name, cols_stmt=cols_stmt)
+        
     def add_indexes(self, columns:List[str]):
         """
         To indexed columns
@@ -668,3 +605,5 @@ class Collection(object):
 
     def __len__(self):
         return self.size
+
+
